#include <gtk/gtk.h>
#include <xmms/xmmsctrl.h>
#include <stdio.h>
#include <stdlib.h>

#include "xpm_bundle.h"

/**** control pad handling routines ****/

GtkWidget *controlpad, *cp_darea_global;
extern const int xmms_session;
int debug = 0;

typedef enum { NONE=0, NO, SO, EA, WE } quadrant;

static const char **cpad_play_xpms[] = {
  cpad_play_xpm, 
  cpad_play_north_xpm, 
  cpad_play_south_xpm, 
  cpad_play_east_xpm, 
  cpad_play_west_xpm
};

static const char **cpad_pause_xpms[] = {
  cpad_pause_xpm, 
  cpad_pause_north_xpm, 
  cpad_pause_south_xpm, 
  cpad_pause_east_xpm, 
  cpad_pause_west_xpm
};

static GdkPixmap *cpad_play_pixmaps[5], *cpad_pause_pixmaps[5];

/* which quadrant of the control pad currently holds the mouse */
static quadrant cp_current_quadrant;

static GdkPixmap **play_or_pause_pixmap(void) {
  if(xmms_remote_is_playing(xmms_session) 
     && !xmms_remote_is_paused(xmms_session)) {
    return(cpad_pause_pixmaps);
  } else {
    return(cpad_play_pixmaps);
  }
}

static void cp_set_pixmap(GtkWidget *cp_darea) {
  gdk_window_set_back_pixmap(cp_darea->window,
                             play_or_pause_pixmap()[cp_current_quadrant],0);
  gdk_window_clear(cp_darea->window);
}
    

static quadrant quadrant_from_coord(float x, float y) {
  if(abs(x-y) < 3) {  /* along the 0,0 - 48,48 diagonal */
    return NONE;
  }
  if(abs((48-x)-y) < 3) {  /* along the 48,0 - 0,48 diagonal */
    return NONE;
  }
  if(x<28 && x > 20 && y<28 && y>20) { /* dead zone */
    return NONE;
  }
  if(x<y) {
    if(x+y>48) {
      return SO;
    } else {
      return WE;
    }
  } else {
    if(x+y>48) {
      return EA;
    } else {
      return NO;
    }
  }
}

static const char *quadrant_tostring(quadrant q) {
  switch (q) {
  case NONE: return "none";
  case NO:   return "north";
  case SO:   return "south";
  case EA:   return "east";
  case WE:   return "west";
  default:   return "?";
  }
}


static gint cp_leave_event( GtkWidget *cp_darea,
                     GdkEvent *event __attribute__((unused))) {
  cp_current_quadrant = NONE;
  cp_set_pixmap(cp_darea);
  return(FALSE);
}

static gint cp_motion_event( GtkWidget *cp_darea,
                      GdkEventMotion  *event) {
  quadrant q = quadrant_from_coord(event->x, event->y);
  if(debug) printf ("whack motion: %f %f: %s\n", 
                    event->x, event->y, quadrant_tostring(q) );
  if(q != cp_current_quadrant) {
    cp_current_quadrant = q;
    cp_set_pixmap(cp_darea);
  }
  return(FALSE);
}

static gint cp_button_release( GtkWidget *cp_darea,
                        GdkEventButton  *event) {
  quadrant q = quadrant_from_coord(event->x, event->y);
  switch(q) {
  case NO:
    if(xmms_remote_is_playing(xmms_session) 
       && !xmms_remote_is_paused(xmms_session)) 
      xmms_remote_pause(xmms_session);
    else 
      xmms_remote_play(xmms_session);
    break;
  case SO:
    xmms_remote_stop(xmms_session);
    break;
  case WE:
    xmms_remote_playlist_prev(xmms_session);
    break;
  case EA:
    xmms_remote_playlist_next(xmms_session);
    break;
  default:
    break;
  }
  gtk_widget_hide(cp_darea->parent); /* controlpad */
  return(FALSE);
}


void build_control_pad(GtkWidget *topwindow) {
  int i;
  GdkBitmap *mask = NULL;
  GtkWidget *cp_darea;
  
  controlpad=gtk_window_new(GTK_WINDOW_POPUP);
  gtk_window_set_policy (GTK_WINDOW(controlpad), FALSE, FALSE, FALSE);
  gtk_widget_set_usize(controlpad, 48, 48);
  cp_darea=gtk_drawing_area_new();

  gtk_widget_set_usize(cp_darea, 48, 48);
  gtk_widget_set_events(cp_darea, GDK_BUTTON_MOTION_MASK 
                        | GDK_BUTTON_RELEASE_MASK
                        | GDK_LEAVE_NOTIFY_MASK
                        | GDK_POINTER_MOTION_MASK );


  gtk_signal_connect(GTK_OBJECT(cp_darea), "motion_notify_event", 
                     GTK_SIGNAL_FUNC(cp_motion_event), NULL);
  gtk_signal_connect(GTK_OBJECT(cp_darea), "leave_notify_event", 
                     GTK_SIGNAL_FUNC(cp_leave_event), NULL);
  gtk_signal_connect(GTK_OBJECT(cp_darea), "button_release_event", 
                     GTK_SIGNAL_FUNC(cp_button_release), NULL);

  gtk_container_add(GTK_CONTAINER(controlpad), cp_darea);
  gtk_widget_realize(controlpad);
  
  for(i=0; i<5; i++) {
    /* TODO: we really only need the control pad transparency mask once. */
    /* this is probably a memory waste. */
    cpad_play_pixmaps[i] = 
      gdk_pixmap_create_from_xpm_d(topwindow->window, &mask, NULL, 
                                   (char **)cpad_play_xpms[i]);
    cpad_pause_pixmaps[i] = 
      gdk_pixmap_create_from_xpm_d(topwindow->window, &mask, NULL, 
                                   (char **)cpad_pause_xpms[i]);
  }
  gtk_widget_shape_combine_mask(controlpad, mask, 0, 0);
  gtk_widget_show(cp_darea);
  /* so I need some side effect of show() to make cp_darea->window
     be non-null for the set_back_pixmap call to succeed */
  gtk_widget_show(controlpad);
  gtk_widget_hide(controlpad);
  gdk_window_set_back_pixmap(cp_darea->window,cpad_pause_pixmaps[NONE],0);
  cp_darea_global = cp_darea;
}

void controlpad_show(void) {
  cp_current_quadrant = NONE;
  cp_set_pixmap(cp_darea_global);
  gtk_window_set_position(GTK_WINDOW(controlpad), GTK_WIN_POS_MOUSE);
  gtk_widget_show(controlpad);
}

void controlpad_hide(void) {
  gtk_widget_hide(controlpad);
}
