#include <X11/Xutil.h>
#include <xmms/xmmsctrl.h>
#include <gdk/gdkx.h>
#include <gtk/gtk.h>
#include <gtk/gtkwindow.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glob.h>
#include <assert.h>
#include <unistd.h>
#include <sys/stat.h>
#include "controlpad.h"
#include "commando.h"
#include "logo.h"

/* the following are command-line parameters. */
boolean want_verbose = FALSE;

/* if you want to use other gdk-pixbuf supported image types, or uppercase... */
#define DEFAULT_APPEND_REGEX ".{jpeg,gif,jpg}"
static const char *append_regex = DEFAULT_APPEND_REGEX;
#define DEFAULT_PREPEND_REGEX "{{F,f}ront,}"
static const char *prepend_regex = DEFAULT_PREPEND_REGEX;

struct commandos commando[] = {
  {
    "Album cover filenames start with this regex, default: '" DEFAULT_PREPEND_REGEX "'",
    "prepend", 'p', required_argument, commando_strdup, &prepend_regex 
  },  
  {
    "Album cover filenames end with this regex, default: '" DEFAULT_APPEND_REGEX "'",
    "append", 'a', required_argument, commando_strdup, &append_regex 
  },  
  { 
    "This help screen", 
    "help", 'h', no_argument, (commando_cb)commando_usage, commando 
  },
  { 
    "Print verbose debugging messages", 
    "verbose", 'v', no_argument, (commando_cb)commando_boolean,  &want_verbose
  },
  { END_COMMANDO }
};

const int xmms_session = 0; /* apparently zero is just fine */

/* find_pic returns a picture name for an mp3 filename.
   I store my album covers in the album's directory, so 
   this works great for me.  */
/* could be manipulated to return a filename embedded in
   an id3 tag, if anyone does that */
/*@null@*/ /* can return null */
static char * find_pic(const char *music_file) {
  char globexp[1024]; /* if it's too small get a life. */
  char *p;            /* pointer to a char within globexp, for manipulation */
  glob_t globber;
  char *retval;       /* the picture filename, allocated storage */
  int zok;            /* zero means ok */

  if(music_file == NULL) {
    /* find some idle icon thing */
    return(NULL);
  }
  if(music_file[0] != '/') {
    fprintf(stderr, 
            "unable to seek picture; playlist must contain full paths:\n");
    fprintf(stderr, "%s\n", music_file);
    return(NULL);
  }
  /* generates a filename of an image to display. */
  strncpy(globexp, music_file, sizeof(globexp));
  p=rindex(globexp, '.');  /* find the . of .mp3 or .leet-media-format */
  if(p == NULL) { return NULL; /* possibly reading from cd. */} 
  strncpy(p, append_regex, sizeof(globexp) - (p-globexp));

  zok = glob(globexp, GLOB_BRACE | GLOB_NOSORT, NULL, &globber);
  if(zok == 0 && globber.gl_pathc > 0) {
    retval = strdup(globber.gl_pathv[0]);
    globfree(&globber);
    return(retval);
  }

  strncpy(globexp, music_file, sizeof(globexp));
  while(p>globexp && (p = rindex(globexp, '/')) != NULL) {
    sprintf(p, "/%s*%s", prepend_regex, append_regex);

    zok = glob(globexp, GLOB_BRACE | GLOB_NOSORT, NULL, &globber);
    if(zok == 0 && globber.gl_pathc > 0) {
      retval = strdup(globber.gl_pathv[0]);
      globfree(&globber);
      return(retval);
    }

    *p='\0';
  }
  return(NULL);
}

static GtkWidget *window,*darea;
static GdkGC *gc;
static GdkPixmap *album_pixmap;
static GdkPixmap *default_pixmap;
static GtkTooltips *tooltips;

static int periodic(/*@unused@*/ void *dummy __attribute__((unused))) {
  int pos, length;
  static int previous_pos, previous_length;
  static int remote_is_dead;

  if(!xmms_remote_is_running(xmms_session)) {
    if(remote_is_dead == 0) {
      gdk_window_set_back_pixmap(darea->window,default_pixmap,0);
      gdk_window_clear(darea->window);
      remote_is_dead = 1;
    }
    return TRUE;
  } else {
    remote_is_dead = 0;
  }

  /* scheme duplicated from xmms' song change plugin */
  pos = xmms_remote_get_playlist_pos(xmms_session);
  length = xmms_remote_get_playlist_time(xmms_session, pos);
  if (pos != previous_pos || length != previous_length) {
    char *pic_fname;
    char *mp3_fname;
    char *song_title;
    GdkPixbuf *pb = NULL, *scaled = NULL;

    previous_pos = pos;
    previous_length = length;

    /* clear out album pixmap, that's kept for mouse movement issues. */
    if(album_pixmap) {
      gdk_pixmap_unref(album_pixmap);
      album_pixmap = NULL;
    }

    mp3_fname = xmms_remote_get_playlist_file(xmms_session, pos);
    song_title = xmms_remote_get_playlist_title(xmms_session, pos);

    /* setup tooltip title, done first in case 
     image stuff fails. */
    gtk_tooltips_set_tip( tooltips, darea, song_title, NULL);

    /* find a file */
    pic_fname = find_pic(mp3_fname);

    gdk_threads_enter();
    if(pic_fname == NULL) {
      gdk_window_set_back_pixmap(darea->window,default_pixmap,0);
      gdk_window_clear(darea->window);
      goto DONE;
    }

    /* fetch picture */
    pb = gdk_pixbuf_new_from_file(pic_fname);
    if(pb==NULL) {
      printf("failed to load '%s'\n", pic_fname);
      goto DONE;
    }
    scaled =gdk_pixbuf_scale_simple (pb, 56, 56, GDK_INTERP_BILINEAR);
    if(scaled==NULL) {
      printf("failed to rescale '%s'\n", pic_fname);
      goto DONE;
    }
    gdk_pixbuf_render_pixmap_and_mask (scaled, &album_pixmap, NULL, 128);
    if(album_pixmap == NULL) {
      printf("failed to convert to pixmap '%s'\n", pic_fname);
      goto DONE;
    } else {
      gdk_window_set_back_pixmap(darea->window,album_pixmap,0);
      gdk_window_clear(darea->window);
    }
    
  DONE: /* handler for failure and success alike */
    if(pb) gdk_pixbuf_unref(pb);
    if(scaled) gdk_pixbuf_unref(scaled);
    if(pic_fname) free(pic_fname);   /* held onto this long for err msgs */
    if(mp3_fname) free(mp3_fname);  
    if(song_title) free(song_title);   
    gdk_threads_leave();
  }
  return TRUE;
}

static void increase_volume(void) {
  int vol = xmms_remote_get_main_volume(xmms_session);
  vol = min(vol+3, 100);
  xmms_remote_set_main_volume(xmms_session, vol);
}
static void decrease_volume(void) {
  int vol = xmms_remote_get_main_volume(xmms_session);
  vol = max(vol-3, 0);
  xmms_remote_set_main_volume(xmms_session, vol);
}

/* taken from wmbiff.c (a piece I wrote too) */
int exists(const char *filename) {
	struct stat st_buf;
	// DMA(DEBUG_INFO, "looking for %s\n", filename);
	if (stat(filename, &st_buf) == 0) {
		if (S_ISREG(st_buf.st_mode)) {
          // DMA(DEBUG_INFO, "found %s\n", filename);
			return (1);
		}
	}
	return (0);
}
/* acts like execvp, with code inspired by it */
/* mustfree */
char *search_path(const char *path, const char *find_me) {
	char *buf;
	const char *p;
	int len, pathlen;
	if (strchr(find_me, '/') != NULL) {
		return (strdup(find_me));
	}
	pathlen = strlen(path);
	len = strlen(find_me) + 1;
	buf = malloc(pathlen + len + 1);
	memcpy(buf + pathlen + 1, find_me, len);
	buf[pathlen] = '/';

	for (p = path; p != NULL; path = p, path++) {
		char *startp;
		p = strchr(path, ':');
		if (p == NULL) {
			/* not found; p should point to the null char at the end */
			startp =
				memcpy(buf + pathlen - strlen(path), path, strlen(path));
		} else if (p == path) {
			/* double colon apparently means try here */
			startp = &buf[pathlen + 1];
		} else {
			/* copy the part between the colons to the buffer */
			startp = memcpy(buf + pathlen - (p - path), path, p - path);
		}
		if (exists(startp)) {
			char *ret = strdup(startp);
			free(buf);
			return (ret);
		}
	}
	free(buf);
	return (NULL);
}


static gint button_press_event( GtkWidget *widget,
                                GdkEventButton *event) {
  assert(darea == widget);
  if (event->type == GDK_BUTTON_PRESS) {
    if(event->button == 4) {
      increase_volume();
    } else if(event->button == 5) {
      decrease_volume();
    } else {
      /* for now, show the control pad on any other button */
      if(xmms_remote_is_running(xmms_session)) {
        controlpad_show();
      }
    }
    return TRUE;
  } else if(event->type == GDK_BUTTON_RELEASE) {
    controlpad_hide();
    return TRUE;
  } else if(event->type == GDK_2BUTTON_PRESS) {
    if(event->button == 1) {
      if(!xmms_remote_is_running(xmms_session)) {
        if(fork() == 0) {
          (void)execlp("xmms", "xmms", NULL);
          exit(EXIT_FAILURE); /* could fail to exec */
        }
      } else {
        /* show the image, full size */
        int pos = xmms_remote_get_playlist_pos(xmms_session);
        char * mp3_fname = xmms_remote_get_playlist_file(xmms_session, pos);
        fprintf(stderr, "about to show\n");
        if(mp3_fname != NULL && mp3_fname[0] == '/') {
          char *pic_fname = find_pic(mp3_fname);
          free(mp3_fname);
          // fprintf(stderr, "found pic %s\n", pic_fname);
          if(pic_fname != NULL) {
            char * runme = search_path(getenv("PATH"), "xv");
            if(runme == NULL) {
              runme = search_path(getenv("PATH"), "xloadimage");
            }
            // fprintf(stderr, "found picture thingy %s\n", runme);
            if(runme) {
              if(fork() == 0) {
                (void)execl(runme, runme, pic_fname, NULL);
                exit(EXIT_FAILURE);
              }
              free(runme);
            } else {
              fprintf(stderr, "did not find xv or xloadimage in your path\n");
            }
          }
          free(pic_fname);
        }
      }
      return TRUE;
    }
  }
  return(FALSE);
}

static void gtkwindow_init(int argc, char **argv);
int main(int argc, char **argv) {
  /* GTK removes stuff from the c/l options in gtk_init - barbarians!
   * We need those for XSetCommand. */
  guint timeout_tag;
  int oargc = argc;
  char **oargv = malloc(argc * sizeof(char*));

  memcpy(oargv, argv, argc * sizeof(char*));

  gtk_init (&argc, &argv);
  (void)commando_parse(argc, argv, commando);
  gtkwindow_init(oargc, oargv);
  free(oargv);
  (void)periodic(NULL);
  timeout_tag = gtk_timeout_add(100, periodic, NULL);
  gtk_main();
  return(0);
}

static void gtkwindow_init(int argc, char **argv) {
  XWMHints mywmhints;
  gdk_threads_leave();
  gdk_threads_enter();
  gdk_rgb_init();
  gtk_widget_set_default_visual(gdk_rgb_get_visual());
  gtk_widget_set_default_colormap(gdk_rgb_get_cmap());

  window=gtk_window_new(GTK_WINDOW_TOPLEVEL);

  gtk_window_set_title(GTK_WINDOW(window), "wmalbum");
  gtk_window_set_policy(GTK_WINDOW(window), FALSE, FALSE, FALSE);
  gtk_widget_set_usize(window, 64, 64);

  gtk_widget_realize(window);
  XSetCommand(GDK_DISPLAY(), GDK_WINDOW_XWINDOW(window->window), argv, argc);

  gc = gdk_gc_new(window->window);
  if(gc == NULL) {
    printf("unable to allocate graphics context\n");
    exit(EXIT_FAILURE);
  }

  darea=gtk_drawing_area_new();
  assert(darea!=NULL);
  gtk_widget_set_usize(darea, 56, 56);
  gtk_widget_set_events(darea, GDK_ENTER_NOTIFY_MASK 
                        | GDK_LEAVE_NOTIFY_MASK
                        | GDK_POINTER_MOTION_MASK
                        | GDK_BUTTON_PRESS_MASK 
                        | GDK_BUTTON_RELEASE_MASK );

  (void)gtk_signal_connect(GTK_OBJECT(darea), "button_press_event", 
                     GTK_SIGNAL_FUNC(button_press_event), NULL);
  (void)gtk_signal_connect(GTK_OBJECT(darea), "button_release_event", 
                     GTK_SIGNAL_FUNC(button_press_event), NULL);

  gtk_container_set_border_width(GTK_CONTAINER(window), 4);
  gtk_container_add(GTK_CONTAINER(window), darea);

  gtk_widget_show(darea);
  gtk_widget_show(window);
  gdk_window_clear(window->window);

  /* trick from http://mail.gnome.org/archives/gtk-list/2000-January/msg00072.html */
  /* by David Raufeisen */
  mywmhints.initial_state = WithdrawnState;
  mywmhints.icon_window = GDK_WINDOW_XWINDOW(darea->window);
  mywmhints.icon_x = 0; 
  mywmhints.icon_y = 0; 
  mywmhints.window_group = GDK_WINDOW_XWINDOW(window->window);
  mywmhints.flags=StateHint | IconWindowHint | IconPositionHint |
    WindowGroupHint;
        
  if(XSetWMHints(GDK_DISPLAY(), GDK_WINDOW_XWINDOW(window->window), &mywmhints) != 1) {
    printf("apparent failure to set window manager hints\n");
  }

  tooltips = gtk_tooltips_new( );
  gtk_tooltips_set_tip( tooltips, darea, "double-click to start xmms", NULL);

  build_control_pad(window);

  default_pixmap = 
    gdk_pixmap_create_from_xpm_d(window->window, NULL, NULL, 
                                 (char **)logo_xpm);

  gdk_threads_leave();
}

